from django.core.exceptions import ObjectDoesNotExist

from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveCharacter

from ledger import app_settings, models
from ledger.errors import LedgerImportError
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def convert_ess_payout(ess: int) -> float:
    return (ess / app_settings.LEDGER_CORP_TAX) * (100 - app_settings.LEDGER_CORP_TAX)


# pylint: disable=import-outside-toplevel
def get_models_and_string():
    if app_settings.LEDGER_MEMBERAUDIT_USE:
        try:
            from memberaudit.models import (
                CharacterMiningLedgerEntry as CharacterMiningLedger,
            )
            from memberaudit.models import CharacterWalletJournalEntry

            return (
                CharacterMiningLedger,
                CharacterWalletJournalEntry,
            )
        except ImportError as exc:
            logger.error("Memberaudit is enabled but not installed")
            raise LedgerImportError("Memberaudit is enabled but not installed") from exc

    from ledger.models.characteraudit import (
        CharacterMiningLedger,
        CharacterWalletJournalEntry,
    )

    return (
        CharacterMiningLedger,
        CharacterWalletJournalEntry,
    )


# pylint: disable=import-outside-toplevel
def get_corp_models_and_string():
    if app_settings.LEDGER_CORPSTATS_TWO:
        try:
            from corpstats.models import CorpMember

            return CorpMember
        except ImportError as exc:
            logger.error("Corpstats is enabled but not installed")
            raise LedgerImportError("Corpstats is enabled but not installed") from exc

    from allianceauth.corputils.models import CorpMember

    return CorpMember


def get_main_character(request, character_id):
    perms = True

    main_char = EveCharacter.objects.select_related(
        "character_ownership",
        "character_ownership__user__profile",
        "character_ownership__user__profile__main_character",
    ).get(character_id=character_id)

    main_char = main_char.character_ownership.user.profile.main_character

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        account_chars = (
            request.user.profile.main_character.character_ownership.user.character_ownerships.all()
        )
        account_char_ids = account_chars.values_list("character_id", flat=True)

        logger.warning(
            "%s Can See %s, requested %s",
            request.user,
            list(visible),
            main_char.id,
        )

        if main_char.id not in account_char_ids:
            perms = False

    if not request.user.has_perm("ledger.basic_access"):
        logger.warning(
            "%s does not have Perm requested, Requested %s", request.user, main_char.id
        )
        perms = False

    return perms, main_char


def get_alts_queryset(main_char):
    try:
        linked_characters = (
            main_char.character_ownership.user.character_ownerships.all().values_list(
                "character_id", flat=True
            )
        )
        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_main_and_alts_all(corporations: list, char_ids=False, corp_members=True):
    """
    Get all mains and their alts from Alliance Auth if they are in Corp

    Args:
    - corporations: corp `list`
    - char_ids: include list
    - corp_members: add Corp Members

    Returns - Dict (Queryset)
    - `Dict`: Mains and Alts QuerySet (EvECharacter or CorpMember)

    Returns - Dict(Queryset) & List
    - `Dict`: Mains and Alts Queryset
    - `List`: Character IDS
    """
    mains = {}
    corpmember = get_corp_models_and_string()

    # pylint: disable=no-member
    users = (
        UserProfile.objects.select_related("main_character")
        .all()
        .order_by("main_character_id")
    )

    for char in users:
        if char.main_character:
            main = char.main_character.character_ownership.user.profile.main_character
            linked_characters = (
                main.character_ownership.user.character_ownerships.all().exclude(
                    character__character_id=main.character_id
                )
            )
            if main.corporation_id in corporations:
                mains[main.character_id] = {
                    "main": main,
                    "alts": [char.character for char in linked_characters],
                }
        else:
            continue

    if corp_members:
        corp = corpmember.objects.select_related("corpstats", "corpstats__corp").filter(
            corpstats__corp__corporation_id__in=corporations
        )

        # Add None Registred Chars to the Ledger
        chars = list(
            set(
                [main["main"].character_id for _, main in mains.items()]
                + [
                    alt.character_id
                    for _, main in mains.items()
                    for alt in main["alts"]
                ]
            )
        )
        for char in corp:
            try:
                char = EveCharacter.objects.get(character_id=char.character_id)
            except EveCharacter.DoesNotExist:
                pass
            if char.character_id not in chars:
                mains[char.character_id] = {"main": char, "alts": []}
    # Sort Names Alphabetic
    mains = sorted(mains.items(), key=lambda item: item[1]["main"].character_name)
    mains = dict(mains)
    if char_ids:
        chars = list(
            set(
                [main["main"].character_id for _, main in mains.items()]
                + [
                    alt.character_id
                    for _, main in mains.items()
                    for alt in main["alts"]
                ]
            )
        )
        return mains, chars
    return mains


def get_corporations(request):  # pylint: disable=unused-argument
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.all().values_list(
        "character_id", flat=True
    )
    chars = EveCharacter.objects.filter(id__in=linked_characters)

    corporations = set()

    for char in chars:
        corporations.add(char.corporation_id)

    return corporations



import sys
from unittest.mock import MagicMock, patch

from corpstats.models import CorpMember as ExpectedCorpMember
from memberaudit.models import (
    CharacterMiningLedgerEntry as ExceptedCharacterMiningLedger,
)
from memberaudit.models import (
    CharacterWalletJournalEntry as ExceptedCharacterWalletJournalEntry,
)

from django.core.exceptions import ObjectDoesNotExist
from django.test import RequestFactory, TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import (
    add_character_to_user,
    add_new_token,
    create_user_from_evecharacter,
)

from ledger.api.helpers import (
    get_alts_queryset,
    get_corp_models_and_string,
    get_main_and_alts_all,
    get_main_character,
    get_models_and_string,
)
from ledger.errors import LedgerImportError
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.api.helpers"


class TestApiHelpers(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
        )
        cls.user4, _ = create_user_from_evecharacter(
            1004,
        )

    def test_get_main_and_alts_all_no_corp(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        char = EveCharacter.objects.get(character_id=1001)
        self.user4.profile.main_character.delete()
        # when
        data = get_main_and_alts_all([2001], corp_members=False)
        # then
        self.assertEqual(data, {1001: {"main": char, "alts": []}})

    def test_get_main_and_alts_all_not_in_chars(self):
        # given
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=EveCorporationInfo.objects.get(corporation_id=2001),
        )
        CorpMember.objects.create(
            character_id=1005, character_name="Gerthd", corpstats=corp_stats
        )
        chars = EveCharacter.objects.filter(character_id__in=[1001, 1004, 1005])
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": []}
        excepted_data = mains
        # when
        data = get_main_and_alts_all([2001], corp_members=True)
        # then
        self.assertEqual(data, excepted_data)

    def test_get_main_and_alts_all_char_in_chars(self):
        # given
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=EveCorporationInfo.objects.get(corporation_id=2001),
        )
        CorpMember.objects.create(
            character_id=1005, character_name="Gerthd", corpstats=corp_stats
        )
        self.user5, _ = create_user_from_evecharacter(
            1005,
        )
        chars = EveCharacter.objects.filter(character_id__in=[1001, 1004, 1005])
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": []}
        excepted_data = mains
        # when
        data = get_main_and_alts_all([2001], corp_members=True)
        # then
        self.assertEqual(data, excepted_data)

    def test_get_main_and_alts_all_char_does_not_exist(self):
        # given
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=EveCorporationInfo.objects.get(corporation_id=2001),
        )
        corp_member = CorpMember.objects.create(
            character_id=1005, character_name="Gerthd", corpstats=corp_stats
        )
        EveCharacter.objects.get(character_id=1005).delete()

        chars = EveCharacter.objects.filter(character_id__in=[1001, 1004])
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": []}
        mains[1005] = {"main": corp_member, "alts": []}

        excepted_data = mains
        # when
        data = get_main_and_alts_all([2001], corp_members=True)
        # then
        self.assertEqual(data, excepted_data)

    def test_get_main_character(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        add_character_to_user(
            self.user, EveCharacter.objects.get(character_id=1004), is_main=False
        )
        data = get_main_character(request, 1004)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (True, character))

    def test_get_main_character_no_permission(self):
        # given
        request = self.factory.get("/")
        request.user = self.user3
        # when
        data = get_main_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (False, character))

    @patch(MODULE_PATH + ".models.CharacterAudit.objects.visible_eve_characters")
    def test_main_char_not_in_account_chars(self, mock_visible):
        # given
        request = self.factory.get("/")
        mock_visible.return_value = []
        request.user = self.user
        # when
        data = get_main_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (True, character))

    def test_get_alts_queryset(self):
        # given
        main_char = EveCharacter.objects.get(character_id=1001)
        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)

    def test_get_alts_queryset_no_linked_characters(self):
        # given
        main_char = MagicMock()
        main_char.character_ownership.user.character_ownerships.all.return_value.values_list.side_effect = (
            ObjectDoesNotExist
        )
        main_char.pk = 90
        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)

    @patch(MODULE_PATH + ".app_settings.LEDGER_MEMBERAUDIT_USE", True)
    def test_get_models_and_string_memberaudit(self):
        CharacterMiningLedger, CharacterWalletJournalEntry = get_models_and_string()

        self.assertIs(CharacterMiningLedger, ExceptedCharacterMiningLedger)
        self.assertIs(CharacterWalletJournalEntry, ExceptedCharacterWalletJournalEntry)

    @patch(MODULE_PATH + ".app_settings.LEDGER_CORPSTATS_TWO", True)
    def test_get_corp_models_and_string(self):
        CorpMember = get_corp_models_and_string()
        self.assertIs(CorpMember, ExpectedCorpMember)


class TestApiHelperCorpStatsImport(TestCase):
    def setUp(self):
        self.original_sys_modules = sys.modules.copy()

    def tearDown(self):
        sys.modules = self.original_sys_modules

    @patch(MODULE_PATH + ".app_settings.LEDGER_CORPSTATS_TWO", True)
    @patch(MODULE_PATH + ".app_settings.LEDGER_MEMBERAUDIT_USE", True)
    @patch(MODULE_PATH + ".logger")
    def test_packages_are_not_installed(self, mock_logger):
        with patch.dict(
            sys.modules,
            {k: None for k in list(sys.modules) if k.startswith("corpstats")},
        ):
            with self.assertRaises(LedgerImportError):
                _ = get_corp_models_and_string()
            mock_logger.error.assert_called()

        with patch.dict(
            sys.modules,
            {k: None for k in list(sys.modules) if k.startswith("memberaudit")},
        ):
            with self.assertRaises(LedgerImportError):
                CharacterMiningLedger, CharacterWalletJournalEntry = (
                    get_models_and_string()
                )
            mock_logger.error.assert_called()



"""
Core View Helper
"""

from decimal import Decimal
from typing import Tuple

from django.core.cache import cache
from django.db.models import Q

from ledger.app_settings import STORAGE_BASE_KEY
from ledger.decorators import custom_cache_timeout
from ledger.hooks import get_extension_logger
from ledger.models.events import Events

logger = get_extension_logger(__name__)


def calculate_ess_stolen(total_amount: int, ess_amount: int) -> Tuple[int, int]:
    total_ess_stolen = 0
    total_ess_gain = 0

    total_ess_gain = ess_amount / total_amount
    total_ess_gain = total_ess_gain * 100

    total_ess_gain_diff = Decimal(66.667) - int(total_ess_gain)
    if abs(total_ess_gain_diff) < Decimal("0.9") or total_ess_gain_diff < 0:
        total_ess_gain = 0
    else:
        total_ess_stolen = ess_amount * (total_ess_gain_diff / 100)
        total_ess_gain = total_ess_gain_diff

    return round(total_ess_stolen), round(total_ess_gain)


def _storage_key(identifier: str) -> str:
    return STORAGE_BASE_KEY + str(identifier)


def get_cache_stale(key_name):
    data = cache.get(key=_storage_key(key_name))
    if not data:
        return False
    return data


def set_cache(data, key_name: str, time: int):
    cache.set(
        key=_storage_key(key_name),
        value=data,
        timeout=custom_cache_timeout(hours=time),
    )


def delete_cache(key_name: str):
    cache.delete(_storage_key(key_name))


def events_filter(entries):
    """
    Filter out all Entries that are in the time of the Event
    """
    # Events to Filter out
    events = Events.objects.all()

    q_objects = []

    # Durchlaufen Sie jedes Event und erstellen Sie das entsprechende Q-Objekt für den Datumsbereich
    for event in events:
        if not event.char_ledger:
            continue
        q_objects.append(Q(date__range=(event.date_start, event.date_end)))

    # Kombinieren Sie alle Q-Objekte mit einer ODER-Verknüpfung
    if q_objects:
        combined_q_object = q_objects[0]
        for q_object in q_objects[1:]:
            combined_q_object |= q_object

        entries = entries.exclude(combined_q_object)
    return entries



from typing import List

from ninja import NinjaAPI

from ledger.api import schema
from ledger.api.helpers import get_alts_queryset, get_main_character
from ledger.api.managers.ledger_manager import JournalProcess
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


# pylint: disable=duplicate-code
class LedgerApiEndpoints:
    tags = ["CharacerLedger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "account/{character_id}/ledger/year/{year}/month/{month}/",
            response={200: List[schema.CharacterLedger], 403: str},
            tags=self.tags,
        )
        def get_character_ledger(request, character_id: int, year: int, month: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, "Permission Denied"

            characters = get_alts_queryset(main)

            # Create the Ledger
            ledger = JournalProcess(characters, year, month)
            output = ledger.character_ledger()

            return output



from django.db.models import DecimalField, F, Q, Sum
from django.db.models.functions import Coalesce

from ledger.api.helpers import convert_ess_payout, get_models_and_string
from ledger.api.managers.billboard_manager import BillboardLedger
from ledger.api.managers.core_manager import (
    LedgerDate,
    LedgerFilter,
    LedgerModels,
    LedgerTotal,
)
from ledger.hooks import get_extension_logger
from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.view_helpers.core import events_filter

logger = get_extension_logger(__name__)

CharacterMiningLedger, CharacterWalletJournalEntry = get_models_and_string()


class JournalProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(self, chars, year, month):
        self.year = year
        self.month = month
        self.chars = chars
        self.corporation_dict = {}
        self.character_dict = {}
        self.summary_total = LedgerTotal()

    def calc_summary_total(self, totals):
        self.summary_total.total_amount += totals.get("total_amount", 0)
        self.summary_total.total_amount_ess += totals.get("total_amount_ess", 0)
        self.summary_total.total_amount_all += totals.get("total_amount_all", 0)
        self.summary_total.total_amount_mining += totals.get("total_amount_mining", 0)
        self.summary_total.total_amount_others += totals.get("total_amount_others", 0)
        self.summary_total.total_amount_costs += totals.get("total_amount_costs", 0)

    def aggregate_journal(self, journal):
        result = journal.aggregate(
            total_amount=Coalesce(Sum(F("amount")), 0, output_field=DecimalField())
        )
        return result["total_amount"]

    def process_corporation_chars(self, corporation_journal):
        # Create a Dict for all Mains(including their alts)
        for _, data in self.chars.items():
            main = data["main"]
            alts = data["alts"]

            chars_mains = [alt.character_id for alt in alts] + [main.character_id]

            total_bounty = 0
            total_ess = 0

            char_name = main.character_name
            char_id = main.character_id

            # Get the Filter Settings
            filters = LedgerFilter(chars_mains)

            total_bounty = self.aggregate_journal(
                corporation_journal.filter(filters.filter_bounty)
            )
            total_ess = self.aggregate_journal(
                corporation_journal.filter(filters.filter_ess)
            )

            summary_amount = total_bounty + total_ess

            if total_bounty or total_ess:
                self.corporation_dict[char_id] = {
                    "main_id": char_id,
                    "main_name": char_name,
                    "alt_names": [],
                    "total_amount": total_bounty,
                    "total_amount_ess": total_ess,
                }

            totals = {
                "total_amount": total_bounty,
                "total_amount_ess": total_ess,
                "total_amount_all": summary_amount,
            }
            # Summary all
            self.calc_summary_total(totals)

    def process_character_chars(
        self, corporation_journal, character_journal, mining_journal
    ):
        """Process the characters for the Journal"""
        chars = [char.character_id for char in self.chars]
        for char in self.chars:
            char_id = char.character_id
            char_name = char.character_name

            # Get the Filter Settings
            filters = LedgerFilter([char_id])

            amounts = {
                "bounty": self.aggregate_journal(
                    character_journal.filter(filters.filter_bounty)
                ),
                "ess": self.aggregate_journal(
                    corporation_journal.filter(filters.filter_ess)
                ),
                "contracts": self.aggregate_journal(
                    character_journal.filter(filters.filter_contract)
                ),
                "transactions": self.aggregate_journal(
                    character_journal.filter(filters.filter_market)
                ),
                "donations": self.aggregate_journal(
                    character_journal.filter(filters.filter_donation).exclude(
                        first_party_id__in=chars
                    )
                ),
                "market_cost": self.aggregate_journal(
                    character_journal.filter(filters.filter_market_cost)
                ),
                "production_cost": self.aggregate_journal(
                    character_journal.filter(filters.filter_production)
                ),
                "mining": mining_journal.filter(filters.filter_mining)
                .values("total", "date")
                .aggregate(
                    total_amount=Coalesce(
                        Sum(F("total")), 0, output_field=DecimalField()
                    )
                ),
            }

            # Convert the ESS Payout for Character
            amounts["ess"] = convert_ess_payout(amounts["ess"])

            total_amount_others = (
                amounts["contracts"] + amounts["transactions"] + amounts["donations"]
            )
            costs_amount = amounts["market_cost"] + amounts["production_cost"]
            summary_amount = (
                amounts["bounty"]
                + amounts["ess"]
                + amounts["mining"]["total_amount"]
                + total_amount_others
            )
            summary_amount -= abs(costs_amount)

            if amounts["bounty"] > 0 or total_amount_others > 0:
                self.character_dict[char_id] = {
                    "main_id": char_id,
                    "main_name": char_name,
                    "total_amount": amounts["bounty"],
                    "total_amount_ess": amounts["ess"],
                    "total_amount_mining": amounts["mining"]["total_amount"],
                    "total_amount_others": total_amount_others,
                    "total_amount_costs": costs_amount,
                }

            totals = {
                "total_amount": amounts["bounty"],
                "total_amount_ess": amounts["ess"],
                "total_amount_all": summary_amount,
                "total_amount_mining": amounts["mining"]["total_amount"],
                "total_amount_others": total_amount_others,
                "total_amount_costs": costs_amount,
            }
            # Summary all
            self.calc_summary_total(totals)

    def character_ledger(self):
        # Get the Character IDs
        chars = [char.character_id for char in self.chars]

        # Get the Filter Settings
        filters = LedgerFilter(chars)

        # Filter the entries for the Year/Month
        filter_date = Q(date__year=self.year)
        if self.month != 0:
            filter_date &= Q(date__month=self.month)

        # Get the Corporation Journal
        corporation_journal = CorporationWalletJournalEntry.objects.filter(
            filters.filter_second_party, filter_date
        ).select_related(
            "division",
            "division__corporation",
            "division__corporation__corporation",
            "first_party",
            "second_party",
        )
        # Exclude Events to avoid wrong stats
        corporation_journal = events_filter(corporation_journal)

        # Get the Mining Journal
        mining_journal = (
            CharacterMiningLedger.objects.filter(filters.filter_mining, filter_date)
        ).annotate_pricing()

        # Filter for the Character Journal
        character_journal = CharacterWalletJournalEntry.objects.filter(
            filters.filter_partys, filter_date
        ).select_related("first_party", "second_party")

        # Create the Dicts for each Character
        self.process_character_chars(
            corporation_journal, character_journal, mining_journal
        )

        # Create Data for Billboard
        date_data = LedgerDate(self.year, self.month)
        models = LedgerModels(
            character_journal=character_journal,
            corporation_journal=corporation_journal,
            mining_journal=mining_journal,
        )

        # Create the Billboard for the Characters
        ledger = BillboardLedger(date_data, models, corp=False)
        billboard_dict = ledger.billboard_char_ledger(chars)

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(self.character_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": self.summary_total.to_dict(),
                "billboard": billboard_dict,
            }
        )

        return output

    def corporation_ledger(self, chars_list: list):
        # Get the Filter Settings
        filters = LedgerFilter(chars_list)

        filter_date = Q(date__year=self.year)
        if not self.month == 0:
            filter_date &= Q(date__month=self.month)

        corporation_journal = CorporationWalletJournalEntry.objects.filter(
            filters.filter_second_party, filter_date
        ).select_related(
            "division",
            "division__corporation",
            "division__corporation__corporation",
            "first_party",
            "second_party",
        )

        # Create the Dicts for each Character
        self.process_corporation_chars(corporation_journal)

        # Create Data for Billboard
        date_data = LedgerDate(self.year, self.month)
        models = LedgerModels(corporation_journal=corporation_journal)

        # Create the Billboard for the Corporation
        ledger = BillboardLedger(date_data, models, corp=True)

        billboard_dict = ledger.billboard_corp_ledger(
            self.corporation_dict, self.summary_total.total_amount
        )

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(self.corporation_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": self.summary_total.to_dict(),
                "billboard": billboard_dict,
            }
        )

        return output



import logging
from typing import List

from ninja import NinjaAPI

from ledger.api import schema
from ledger.api.helpers import get_corporations, get_main_and_alts_all
from ledger.api.managers.ledger_manager import JournalProcess
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class LedgerApiEndpoints:
    tags = ["CorporationLedger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/ledger/year/{year}/month/{month}/",
            response={200: List[schema.CorporationLedger], 403: str},
            tags=self.tags,
        )
        def get_corporation_ledger(request, corporation_id: int, year: int, month: int):
            perms = request.user.has_perm("ledger.basic_access")

            corporations = [corporation_id]

            if corporation_id == 0:
                corporations = get_corporations(request)

            if not perms:
                logging.error("Permission Denied for %s to view wallets!", request.user)
                return 403, "Permission Denied!"

            # filters &= Q(ref_type=ref_type)
            characters, chars_list = get_main_and_alts_all(corporations, char_ids=True)

            # Create the Ledger
            ledger = JournalProcess(characters, year, month)
            output = ledger.corporation_ledger(chars_list)

            return output



from datetime import timedelta
from unittest.mock import MagicMock, call, patch

from django.test import TestCase
from django.utils import timezone
from esi.errors import TokenExpiredError

from app_utils.testing import create_user_from_evecharacter

from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.tasks import (
    update_all_characters,
    update_all_corps,
    update_char_mining_ledger,
    update_char_wallet,
    update_character,
    update_corp,
    update_corp_wallet,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.tasks"


class TestTasks(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.char_audit_admin_access",
                "ledger.corp_audit_admin_access",
            ],
        )
        cls.token = cls.user.token_set.first()
        cls.corporation = cls.character_ownership.character.corporation

    @patch(MODULE_PATH + ".update_character.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_characters(self, mock_logger, mock_update_character):
        # given
        characters_count = CharacterAudit.objects.count()
        # when
        update_all_characters()
        # then
        self.assertEqual(mock_update_character.call_count, characters_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Char Audit Updates", characters_count
        )

    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character(self, mock_select_related):
        # given
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now()
        mock_character.last_update_wallet = timezone.now()
        mock_character.character = MagicMock()
        mock_character.character.character_id = self.token.character_id
        mock_character.character.character_name = self.token.character_name

        mock_filter = MagicMock()
        mock_filter.first.return_value = mock_character
        mock_select_related.return_value.filter.return_value = mock_filter
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".update_character_mining")
    def test_update_character_mining(self, mock_char_mining):
        # given
        expected_return_value = f"Finished Mining for: {self.token.character_name}"
        mock_char_mining.return_value = expected_return_value
        # when
        result = update_char_mining_ledger(self.token.character_id)
        # then
        self.assertTrue(mock_char_mining.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_character_wallet")
    def test_update_character_wallet(self, mock_char_wallet):
        # given
        expected_return_value = (
            f"Finished wallet transactions for: {self.token.character_name}"
        )
        mock_char_wallet.return_value = expected_return_value
        # when
        result = update_char_wallet(self.token.character_id)
        # then
        self.assertTrue(mock_char_wallet.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_char_mining_ledger.si")
    @patch(MODULE_PATH + ".update_char_wallet.si")
    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_from_token(
        self,
        mock_check_char,
        mock_update_or_create,
        mock_get_character_by_id,
        mock_get_token,
        mock_char_wallet,
        mock_char_mining,
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = True
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now() - timedelta(days=1)
        mock_character.last_update_wallet = timezone.now() - timedelta(days=1)
        mock_update_or_create.return_value = (mock_character, True)
        # when
        update_character(self.token.character_id)
        # then
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()
        mock_get_character_by_id.assert_called_once_with(mock_token.character_id)
        mock_update_or_create.assert_called_once()
        self.assertTrue(mock_char_wallet.called)
        self.assertTrue(mock_char_mining.called)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_expired_token(
        self, mock_check_char, mock_update_or_create, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.side_effect = TokenExpiredError
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_update_or_create.return_value = (mock_character, True)
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_no_access_token(
        self, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = False
        mock_get_token.return_value = mock_token
        # when
        result = update_character(self.token.character_id)
        # then
        mock_token.valid_access_token.assert_called_once()
        self.assertFalse(result)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_token_no_token(
        self, mock_logger, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_get_token.return_value = False
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_logger.info.assert_called_once_with("No Tokens for %s", 1001)

    @patch(MODULE_PATH + ".update_corp.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_cors(self, mock_logger, mock_update_character):
        # given
        corporation_count = CorporationAudit.objects.count()
        # when
        update_all_corps()
        # then
        self.assertEqual(mock_update_character.call_count, corporation_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Corp Audit Updates", corporation_count
        )

    @patch(MODULE_PATH + ".update_corp_wallet.si")
    def test_update_corp(self, mock_corp_wallet):
        # when
        update_corp(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet.called)

    @patch(MODULE_PATH + ".update_corp_wallet_division")
    def test_update_corp_wallet(self, mock_corp_wallet_division):
        # given
        expected_return_value = (
            f"Finished wallet divs for: {self.corporation.corporation_name}"
        )
        mock_corp_wallet_division.return_value = expected_return_value
        # when
        result = update_corp_wallet(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet_division.called)
        self.assertEqual(expected_return_value, result)



from http import HTTPStatus

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import RequestFactory, TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserMainFactory

from ledger.models.general import General
from ledger.views.pve import ledger_index, ratting_char_index, ratting_index


class TestViews(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user = UserMainFactory(
            permissions=[
                "ledger.basic_access",
            ]
        )

    def test_view(self):
        request = self.factory.get(reverse("ledger:index"))
        request.user = self.user
        response = ledger_index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_char_ledger_view(self):
        request = self.factory.get(reverse("ledger:ledger_char_index"))
        request.user = self.user
        response = ratting_char_index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_corp_ledger_view(self):
        request = self.factory.get(reverse("ledger:ledger_corp_index"))
        request.user = self.user
        response = ratting_index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)



from dataclasses import asdict, dataclass, field
from datetime import datetime

from django.db.models import Q

from ledger import app_settings


@dataclass
class LedgerDataCore:
    total_bounty: int = 0
    total_ess_payout: int = 0
    total_mining: int = 0
    total_miscellaneous: int = 0
    total_isk: int = 0


@dataclass
class LedgerData(LedgerDataCore):
    total_cost: int = 0
    total_production_cost: int = 0
    total_market_cost: int = 0


class LedgerModels:
    """LedgerModels class to store the models."""

    def __init__(
        self, character_journal=None, corporation_journal=None, mining_journal=None
    ):
        self.char_journal = character_journal
        self.corp_journal = corporation_journal
        self.mining_journal = mining_journal


@dataclass
class LedgerDate:
    year: int
    month: int
    monthly: bool = field(init=False)
    current_date: datetime = field(default_factory=datetime.now, init=False)
    range_data: int = field(init=False)
    day_checks: list = field(init=False)

    def __post_init__(self):
        self.monthly = self.month == 0
        self.range_data = 12 if self.monthly else 31
        self.day_checks = list(range(1, self.range_data + 1))


@dataclass
class LedgerTotal:
    total_amount: int = 0
    total_amount_ess: int = 0
    total_amount_all: int = 0
    total_amount_mining: int = 0
    total_amount_others: int = 0
    total_amount_costs: int = 0

    def to_dict(self):
        return asdict(self)


@dataclass
class LedgerFilterCore:
    char_id: list

    filter_first_party: Q = field(init=False)
    filter_second_party: Q = field(init=False)
    filter_partys: Q = field(init=False)
    filter_bounty: Q = field(init=False)
    filter_ess: Q = field(init=False)
    filter_mining: Q = field(init=False)

    def __post_init__(self):
        self.filter_first_party = Q(first_party_id__in=self.char_id)
        self.filter_second_party = Q(second_party_id__in=self.char_id)
        self.filter_partys = self.filter_first_party | self.filter_second_party
        self.filter_bounty = self.filter_second_party & Q(ref_type="bounty_prizes")
        self.filter_ess = self.filter_second_party & Q(ref_type="ess_escrow_transfer")
        self.filter_mining = (
            Q(character__eve_character__character_id__in=self.char_id)
            if app_settings.LEDGER_MEMBERAUDIT_USE
            else Q(character__character__character_id__in=self.char_id)
        )


@dataclass
class LedgerFilterCost(LedgerFilterCore):
    """LedgerFilter class to store the filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.filter_market_cost = self.filter_partys & Q(
            ref_type__in=[
                "market_escrow",
                "transaction_tax",
                "market_provider_tax",
                "brokers_fee",
            ],
            amount__lt=0,
        )
        self.filter_production = self.filter_partys & Q(
            ref_type__in=["industry_job_tax", "manufacturing"]
        )


@dataclass
class LedgerFilterTrading(LedgerFilterCost):
    """LedgerFilter class to store the filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.filter_market = self.filter_partys & Q(ref_type="market_transaction")
        self.filter_contract = self.filter_partys & Q(
            ref_type__in=[
                "contract_price_payment_corp",
                "contract_reward",
                "contract_price",
            ],
            amount__gt=0,
        )
        self.filter_donation = self.filter_partys & Q(
            ref_type="player_donation", amount__gt=0
        )


@dataclass
class LedgerFilter(LedgerFilterTrading):
    """LedgerFilterAll class to store all filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.char_id = char_id



import math
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal

from ledger import app_settings
from ledger.api.managers.core_manager import LedgerData, LedgerDate, LedgerModels


@dataclass
class LedgerSum:
    sum_amount: list = field(default_factory=lambda: ["Ratting"])
    sum_amount_ess: list = field(default_factory=lambda: ["ESS Payout"])
    sum_amount_misc: list = field(default_factory=lambda: ["Miscellaneous"])
    sum_amount_mining: list = field(default_factory=lambda: ["Mining"])
    total_sum: int = None


class BillboardLedger:
    def __init__(self, date_data: LedgerDate, models: LedgerModels, corp=False):
        self.is_corp = corp
        self.data = LedgerData()
        self.models = models
        self.date = date_data
        self.sum = LedgerSum()
        self.billboard_dict = {
            "walletcharts": [],
            "charts": [],
            "rattingbar": [],
            "workflowgauge": [],
        }
        self.date_billboard = ["x"]

    def aggregate_corp(self, journal, range_):
        """Aggregate the journal entries for the corporation."""
        total_bounty = 0
        total_ess_payout = 0
        for entry in journal:
            if (self.date.month == 0 and entry["date"].month == range_) or (
                self.date.month != 0 and entry["date"].day == range_
            ):
                if self.is_corp:
                    if entry["ref_type"] == "bounty_prizes":
                        total_bounty += entry["amount"]
                if entry["ref_type"] == "ess_escrow_transfer":
                    if self.is_corp:
                        total_ess_payout += entry["amount"]
                    else:
                        total_ess_payout += (
                            entry["amount"] / app_settings.LEDGER_CORP_TAX
                        ) * (100 - app_settings.LEDGER_CORP_TAX)
        return total_bounty, total_ess_payout

    def aggregate_char(self, journal, range_):
        """Aggregate the journal entries for the character."""
        total_bounty = 0
        total_miscellaneous = 0
        total_isk = 0
        total_cost = 0
        total_market_cost = 0
        total_production_cost = 0

        for entry in journal:
            if (self.date.month == 0 and entry["date"].month == range_) or (
                self.date.month != 0 and entry["date"].day == range_
            ):
                # Bounty Filter
                if entry["ref_type"] == "bounty_prizes":
                    total_bounty += entry["amount"]
                # Misc Filter
                if (
                    entry["ref_type"]
                    in [
                        "market_transaction",
                        "contract_price_payment_corp",
                        "contract_reward",
                        "contract_price",
                    ]
                    and entry["amount"] > 0
                ):
                    total_miscellaneous += entry["amount"]
                # Donation Filter
                if (entry["ref_type"] == "player_donation") and entry["amount"] > 0:
                    if entry["first_party_id"] not in self.chars:
                        total_miscellaneous += entry["amount"]
                # Total ISK
                if entry["amount"] > 0:
                    total_isk += entry["amount"]
                else:
                    total_cost += entry["amount"]
                # Total Market
                if entry["ref_type"] in [
                    "market_escrow",
                    "transaction_tax",
                    "market_provider_tax",
                    "brokers_fee",
                ]:
                    if entry["amount"] < 0:
                        total_market_cost += entry["amount"]
                # Production Cost
                if entry["ref_type"] in ["industry_job_tax", "manufacturing"]:
                    total_production_cost += entry["amount"]
        return (
            total_bounty,
            total_miscellaneous,
            total_isk,
            total_cost,
            total_market_cost,
            total_production_cost,
        )

    def aggregate_mining(self, journal, range_):
        """Aggregate the journal entries for the mining."""
        total_mining = 0
        for entry in journal:
            if (self.date.month == 0 and entry["date"].month == range_) or (
                self.date.month != 0 and entry["date"].day == range_
            ):
                total_mining += entry["total"] if entry["total"] else 0
        return total_mining

    # pylint: disable=too-many-branches
    def process_day(
        self, range_, corp_journal=None, char_journal=None, mining_journal=None
    ):
        """Process the day for the journal entries."""
        date = datetime.now()

        if self.date.monthly:
            date = date.replace(day=1)

        if corp_journal:
            total_bounty, total_ess = self.aggregate_corp(corp_journal, range_)
            self.data.total_bounty = total_bounty
            self.data.total_ess_payout = total_ess
        if char_journal:
            (
                total_bounty,
                total_miscellaneous,
                total_isk,
                total_cost,
                total_market_cost,
                total_production_cost,
            ) = self.aggregate_char(char_journal, range_)

            self.data.total_bounty = total_bounty
            self.data.total_miscellaneous = total_miscellaneous
            # Wallet Charts
            self.data.total_isk += total_isk
            self.data.total_cost += total_cost
            self.data.total_market_cost += total_market_cost
            self.data.total_production_cost += total_production_cost

        if mining_journal:
            total_mining = self.aggregate_mining(mining_journal, range_)
            self.data.total_mining = total_mining

        # Add the totals to the respective lists
        self.sum.sum_amount.append(int(self.data.total_bounty))
        self.sum.sum_amount_ess.append(int(self.data.total_ess_payout))
        if not self.is_corp:
            self.sum.sum_amount_misc.append(int(self.data.total_miscellaneous))
            self.sum.sum_amount_mining.append(int(self.data.total_mining))
        # Add the date to the date list
        try:
            self.date_billboard.append(
                date.replace(day=range_).strftime("%Y-%m-%d")
                if not self.date.monthly
                else date.replace(month=range_).strftime("%Y-%m")
            )
        except ValueError:
            pass

    def calculate_total_sum(self):
        self.sum.total_sum = sum(
            sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
            for lst in [
                self.sum.sum_amount,
                self.sum.sum_amount_ess,
                self.sum.sum_amount_misc,
                self.sum.sum_amount_mining,
            ]
        )

    def calculate_percentages(self):
        percentages = [
            (
                (
                    sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
                    / self.sum.total_sum
                    * 100
                )
                if self.sum.total_sum > 0
                else 0
            )
            for lst in [
                self.sum.sum_amount,
                self.sum.sum_amount_ess,
                self.sum.sum_amount_misc,
                self.sum.sum_amount_mining,
            ]
        ]
        return [math.floor(perc * 10) / 10 for perc in percentages]

    def generate_gauge_data(self, rounded_percentages):
        self.billboard_dict["workflowgauge"] = (
            [
                ["Ratting", rounded_percentages[0]],
                ["ESS Payout", rounded_percentages[1]],
                ["Miscellaneous", rounded_percentages[2]],
                ["Mining", rounded_percentages[3]],
            ]
            if sum(rounded_percentages)
            else None
        )

    def generate_wallet_ratting_bar(self):
        self.billboard_dict["rattingbar"] = (
            (
                [
                    self.date_billboard,
                    self.sum.sum_amount,
                    self.sum.sum_amount_ess,
                    self.sum.sum_amount_misc,
                    self.sum.sum_amount_mining,
                ]
            )
            if self.sum.total_sum
            else None
        )

    def generate_wallet_charts_data(self):
        # Daten für die Wallet-Charts vorbereiten
        misc_cost = abs(
            self.data.total_cost
            - self.data.total_production_cost
            - self.data.total_market_cost
        )
        wallet_chart_data = [
            # Earns
            ["Income", int(self.data.total_isk)],
            # ["Ratting", int(total_sum_ratting)],  ["Mining", int(total_sum_mining)], ["Misc.", int(total_sum_misc)],
            # Costs
            ["Market Cost", int(abs(self.data.total_market_cost))],
            ["Production Cost", int(abs(self.data.total_production_cost))],
            ["Misc. Cost", int(misc_cost)],
        ]

        self.billboard_dict["walletcharts"] = (
            wallet_chart_data
            if any(item[1] != 0 for item in wallet_chart_data)
            else None
        )

    # Generate Charts Billboard for Corporation Wallets
    def generate_wallet_corps_data(self, corporation_dict, summary_dict_all):
        for entry in corporation_dict.values():
            total_amount_entry = entry["total_amount"]
            percentage = (total_amount_entry / summary_dict_all) * 100
            name = entry["main_name"]
            self.billboard_dict["charts"].append([name, percentage])

        self.billboard_dict["charts"] = (
            sorted(self.billboard_dict["charts"], key=lambda x: x[0])
            if self.billboard_dict["charts"]
            else None
        )

    # Create the Billboard Char Ledger
    def billboard_char_ledger(self, chars: list):
        self.chars = chars
        corp_journal_values = self.models.corp_journal.values(
            "amount", "ref_type", "first_party_id", "second_party_id", "date"
        )
        char_journal_values = self.models.char_journal.values(
            "amount", "ref_type", "first_party_id", "second_party_id", "date"
        )
        mining_journal_values = self.models.mining_journal.values("total", "date")
        for day in self.date.day_checks:
            self.process_day(
                day, corp_journal_values, char_journal_values, mining_journal_values
            )

        self.calculate_total_sum()

        rounded_percentages = self.calculate_percentages()
        self.generate_gauge_data(rounded_percentages)

        self.generate_wallet_ratting_bar()

        self.generate_wallet_charts_data()

        return self.billboard_dict

    # Create the Billboard Corp Ledger
    def billboard_corp_ledger(self, corporation_dict, summary_dict_all):
        corp_journal_values = self.models.corp_journal.values(
            "amount", "ref_type", "first_party_id", "second_party_id", "date"
        )
        for day in self.date.day_checks:
            self.process_day(day, corp_journal_values)

        self.calculate_total_sum()

        self.generate_wallet_ratting_bar()

        self.generate_wallet_corps_data(corporation_dict, summary_dict_all)

        return self.billboard_dict



from dataclasses import dataclass
from datetime import datetime

from django.db.models import DecimalField, F, Q, Sum
from django.db.models.functions import Coalesce

from ledger.api.helpers import (
    convert_ess_payout,
    get_alts_queryset,
    get_main_and_alts_all,
    get_main_character,
    get_models_and_string,
)
from ledger.api.managers.core_manager import LedgerFilter
from ledger.hooks import get_extension_logger
from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.view_helpers.core import events_filter

CharacterMiningLedger, CharacterWalletJournalEntry = get_models_and_string()

logger = get_extension_logger(__name__)


@dataclass
class TemplateData:
    """TemplateData class to store the data."""

    request: any
    request_id: int
    year: int
    month: int
    current_date: datetime = datetime.now()

    def __post_init__(self):
        self.current_date = self.current_date.replace(year=self.year)
        if self.month != 0:
            self.current_date = self.current_date.replace(month=self.month)


@dataclass
class TemplateTotalCore:
    """TemplateTotalCore class to store the core data."""

    bounty: int = 0
    ess: int = 0
    mining: int = 0
    contract: int = 0
    transaction: int = 0
    donation: int = 0
    production_cost: int = 0
    market_cost: int = 0


@dataclass
class TemplateTotalDay:
    """TemplateTotalDay class to store the daily data."""

    bounty_day: int = 0
    ess_day: int = 0
    mining_day: int = 0
    contract_day: int = 0
    transaction_day: int = 0
    donation_day: int = 0
    production_cost_day: int = 0
    market_cost_day: int = 0


@dataclass
class TemplateTotalHour:
    """TemplateTotalHour class to store the hourly data."""

    bounty_hour: int = 0
    ess_hour: int = 0
    mining_hour: int = 0
    contract_hour: int = 0
    transaction_hour: int = 0
    donation_hour: int = 0
    production_cost_hour: int = 0
    market_cost_hour: int = 0


@dataclass
class TemplateTotal(TemplateTotalCore, TemplateTotalDay, TemplateTotalHour):
    """TemplateTotal class to store the data."""

    def to_dict(self):
        return {
            "bounty": {
                "total_amount": self.bounty,
                "total_amount_day": self.bounty_day,
                "total_amount_hour": self.bounty_hour,
            },
            "ess": {
                "total_amount": self.ess,
                "total_amount_day": self.ess_day,
                "total_amount_hour": self.ess_hour,
            },
            "mining": {
                "total_amount": self.mining,
                "total_amount_day": self.mining_day,
            },
            "contract": {
                "total_amount": self.contract,
                "total_amount_day": self.contract_day,
                "total_amount_hour": self.contract_hour,
            },
            "transaction": {
                "total_amount": self.transaction,
                "total_amount_day": self.transaction_day,
                "total_amount_hour": self.transaction_hour,
            },
            "donation": {
                "total_amount": self.donation,
                "total_amount_day": self.donation_day,
                "total_amount_hour": self.donation_hour,
            },
            "production_cost": {
                "total_amount": self.production_cost,
                "total_amount_day": self.production_cost_day,
                "total_amount_hour": self.production_cost_hour,
            },
            "market_cost": {
                "total_amount": self.market_cost,
                "total_amount_day": self.market_cost_day,
                "total_amount_hour": self.market_cost_hour,
            },
        }


class TemplateProcess:
    """TemplateProcess class to process the data."""

    def __init__(self, chars, data: TemplateData, show_year=False):
        self.data = data
        self.chars = chars
        self.show_year = show_year
        self.template_dict = {}

    # pylint: disable=duplicate-code
    def character_template(self):
        """
        Create the character template.
        return: dict
        """
        # Create Char List
        chars = [char.character_id for char in self.chars]

        filters = LedgerFilter(chars)

        filter_date = Q(date__year=self.data.year)
        if not self.data.month == 0:
            filter_date &= Q(date__month=self.data.month)

        # Filter the entries for the current day/month
        character_journal = CharacterWalletJournalEntry.objects.filter(
            filters.filter_partys, filter_date
        ).select_related("first_party", "second_party")

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(
                filters.filter_partys, filter_date
            )
            .select_related("first_party", "second_party")
            .order_by("-date")
        )

        # Exclude Events to avoid wrong stats
        corporation_journal = events_filter(corporation_journal)
        mining_journal = (
            CharacterMiningLedger.objects.filter(filters.filter_mining, filter_date)
        ).annotate_pricing()

        self._process_characters(character_journal, corporation_journal, mining_journal)
        return self.template_dict

    def corporation_template(self):
        """
        Create the corporation template.
        return: dict
        """

        mains, chars = get_main_and_alts_all(self.chars, char_ids=True)

        filters = LedgerFilter(chars)
        filter_date = Q(date__year=self.data.year)
        if not self.data.month == 0:
            filter_date &= Q(date__month=self.data.month)

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(
                filters.filter_second_party, filter_date
            )
            .select_related("first_party", "second_party", "division")
            .values("amount", "date", "second_party_id", "ref_type")
            .order_by("-date")
        )

        if self.show_year:
            mains_data = mains
        else:
            mains_data = {self.data.request_id: mains.get(self.data.request_id, None)}

        # Exclude Events to avoid wrong stats
        corporation_journal = events_filter(corporation_journal)

        self._process_corporation(corporation_journal, mains_data)
        return self.template_dict

    # Process the character
    def _process_characters(
        self, character_journal, corporation_journal, mining_journal
    ):
        """Process the characters."""
        _, main = get_main_character(self.data.request, self.data.request_id)

        alts = get_alts_queryset(main)

        chars_list = [char.character_id for char in alts]

        total_amounts = TemplateTotal().to_dict()

        models = character_journal, corporation_journal, mining_journal

        for char in self.chars:
            amounts = self._process_amounts_char(char, models, chars_list)

            # Add amounts to total_amounts
            for key, value in amounts.items():
                for sub_key, sub_value in value.items():
                    total_amounts[key][sub_key] += sub_value

            self._update_template_dict(char)
        self._generate_amounts_dict(total_amounts)

    # Process the corporation
    def _process_corporation(self, corporation_journal, mains_data):
        """Process the characters."""

        total_amounts = {
            "bounty": {
                "total_amount": 0,
                "total_amount_day": 0,
                "total_amount_hour": 0,
            },
            "ess": {"total_amount": 0, "total_amount_day": 0, "total_amount_hour": 0},
        }

        for _, main_data in mains_data.items():
            main = main_data["main"]
            alts = main_data["alts"]

            # Each Chars from a Main Character
            chars = [alt.character_id for alt in alts] + [main.character_id]

            filters = LedgerFilter(chars)

            amounts = {
                # Calculate Income
                "bounty": self._aggregate_journal(
                    corporation_journal.filter(filters.filter_bounty)
                ),
                "ess": self._aggregate_journal(
                    corporation_journal.filter(filters.filter_ess)
                ),
            }

            # Add amounts to total_amounts
            for key, value in amounts.items():
                for sub_key, sub_value in value.items():
                    total_amounts[key][sub_key] += sub_value

            self._update_template_dict(main)

        self._generate_amounts_dict(total_amounts)

    # Aggregate Journal
    def _aggregate_journal(self, journal):
        result = journal.aggregate(
            total_amount=Coalesce(Sum(F("amount")), 0, output_field=DecimalField()),
            total_amount_day=Coalesce(
                Sum(F("amount"), filter=Q(date__day=self.data.current_date.day)),
                0,
                output_field=DecimalField(),
            ),
            total_amount_hour=Coalesce(
                Sum(
                    F("amount"),
                    filter=Q(
                        date__day=self.data.current_date.day,
                        date__hour=self.data.current_date.hour,
                    ),
                ),
                0,
                output_field=DecimalField(),
            ),
        )
        if self.data.month == 0:
            result["total_amount_day"] = 0
        return result

    # Update Core Dict
    def _update_template_dict(self, char):
        main_name = char.character_name if not self.show_year else "Summary"
        main_id = char.character_id if not self.show_year else 0
        date = (
            str(self.data.current_date.year)
            if self.data.month == 0
            else self.data.current_date.strftime("%B")
        )

        self.template_dict.update(
            {
                "main_name": main_name,
                "main_id": main_id,
                "date": date,
            }
        )

    # Add Amounts to Dict
    def _generate_amounts_dict(self, amounts):
        """Generate the amounts dictionary."""
        current_day = 365 if self.data.month == 0 else self.data.current_date.day
        # Calculate the total sum
        total_sum = sum(amounts[key]["total_amount"] for key in amounts)

        self.template_dict.update(
            {
                key: {
                    sub_key: value
                    for sub_key, value in {
                        "total_amount": round(amounts[key]["total_amount"], 2),
                        "total_amount_day": (
                            round(amounts[key]["total_amount_day"], 2)
                        ),
                        "total_amount_hour": (
                            round(amounts[key]["total_amount_hour"], 2)
                            if key != "mining"
                            # Bad Fix but it works..
                            else round(amounts[key]["total_amount_day"], 2)
                        ),
                        "average_day": round(
                            amounts[key]["total_amount"] / current_day, 2
                        ),
                        "average_hour": round(
                            (amounts[key]["total_amount"] / current_day) / 24, 2
                        ),
                    }.items()
                    if value != 0
                }
                for key in amounts
            }
        )
        self.template_dict["summary"] = {
            "total_amount": round(total_sum, 2),
            "total_amount_day": round(total_sum / current_day, 2),
            "total_amount_hour": round((total_sum / current_day) / 24, 2),
        }

    # Genereate Amounts for each Char
    def _process_amounts_char(self, char, models, chars_list):
        char_id = char.character_id

        # Create the filters
        filters = LedgerFilter([char_id])

        # Set the models
        character_journal, corporation_journal, mining_journal = models

        # Calculate the amounts
        amounts = {
            # Calculate Income
            "bounty": self._aggregate_journal(
                character_journal.filter(filters.filter_bounty)
            ),
            "ess": self._aggregate_journal(
                corporation_journal.filter(filters.filter_ess)
            ),
            "mining": mining_journal.filter(filters.filter_mining)
            .values("total", "date")
            .aggregate(
                total_amount=Coalesce(Sum(F("total")), 0, output_field=DecimalField()),
                total_amount_day=Coalesce(
                    Sum(F("total"), filter=Q(date__day=self.data.current_date.day)),
                    0,
                    output_field=DecimalField(),
                ),
            ),
            # Calculate Trading
            "contract": self._aggregate_journal(
                character_journal.filter(filters.filter_contract)
            ),
            "transaction": self._aggregate_journal(
                character_journal.filter(filters.filter_market)
            ),
            "donation": self._aggregate_journal(
                character_journal.filter(filters.filter_donation).exclude(
                    first_party_id__in=chars_list
                )
            ),
            # Calculate Costs
            "production_cost": self._aggregate_journal(
                character_journal.filter(filters.filter_production)
            ),
            "market_cost": self._aggregate_journal(
                character_journal.filter(filters.filter_market_cost)
            ),
        }
        # Convert ESS Payout for Character Ledger
        amounts["ess"]["total_amount"] = convert_ess_payout(
            amounts["ess"]["total_amount"]
        )

        return amounts



"""PvE Views"""

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

# Voices of War
from ledger.app_settings import LEDGER_MEMBERAUDIT_USE
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@login_required
@permission_required("ledger.basic_access")
def ledger_index(request):
    context = {
        "memberaudit": LEDGER_MEMBERAUDIT_USE,
    }
    return render(request, "ledger/index.html", context=context)


@login_required
@permission_required("ledger.basic_access")
def ratting_index(request):
    context = {
        "memberaudit": LEDGER_MEMBERAUDIT_USE,
    }
    return render(request, "ledger/corpledger/corp_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access")
def ratting_char_index(request):
    context = {
        "memberaudit": LEDGER_MEMBERAUDIT_USE,
    }
    return render(request, "ledger/charledger/char_ledger.html", context=context)



from typing import List

from ninja import NinjaAPI

from django.shortcuts import render

from ledger.api import schema
from ledger.api.helpers import get_corporations
from ledger.api.managers.template_manager import TemplateData, TemplateProcess
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
class LedgerTemplateApiEndpoints:
    tags = ["CorporationLedgerTemplate"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "corporation/{main_id}/ledger/template/year/{year}/month/{month}/",
            response={200: List[schema.CharacterLedgerTemplate], 403: str},
            tags=self.tags,
        )
        def get_corporation_ledger_template(
            request, main_id: int, year: int, month: int
        ):
            perms = request.user.has_perm("ledger.basic_access")

            overall_mode = main_id == 0

            if not perms:
                return 403, "Permission Denied"

            corporations = get_corporations(request)

            # Create the Ledger
            ledger_data = TemplateData(request, main_id, year, month)
            ledger = TemplateProcess(corporations, ledger_data, overall_mode)
            context = {"character": ledger.corporation_template()}

            return render(
                request, "ledger/modals/pve/view_character_content.html", context
            )



from unittest.mock import Mock, patch

from memberaudit.models import Character

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_memberaudit import load_memberaudit
from ledger.views.corporation.corp_audit import add_corp

MODULE_PATH = "ledger.views.corporation.corp_audit"


class CharAuditTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_memberaudit()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.corp_audit_admin_access",
            ],
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_corp")
    def test_add_corp(self, mock_update_corp, mock_messages):
        self.client.force_login(self.user)
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        token.corporation_id = self.character_ownership.character.corporation_id
        token.corporation_name = self.character_ownership.character.corporation_name
        token.corporation_ticker = self.character_ownership.character.corporation_ticker
        request = self.factory.get(reverse("ledger:ledger_add_corp"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_corp.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_update_corp.apply_async.called)
        self.assertTrue(
            CorporationAudit.objects.get(
                corporation=self.character_ownership.character.corporation
            )
        )
